@@ -1,7 +1,7 @@
 /*
  *   cddbd - CD Database Protocol Server
  *
- *   Copyright (C) 1996  Steve Scherf
+ *   Copyright (C) 1996-1997  Steve Scherf
  *   Email: steve@moonsoft.com
  *   Moondog Software Productions - makers of fine public domain software.
  *
@@ -22,9 +22,10 @@
  */
 
 #ifndef LINT
-static char *_inet_c_ident_ = "@(#)$Id: inet.c,v 1.5 1996/12/22 01:48:54 steve Exp $";
+static char *_inet_c_ident_ = "@(#)$Id: inet.c,v 1.11 2001/09/11 23:21:34 joerg78 Exp $";
 #endif
 
+#include <stdlib.h>
 #include <sys/types.h>
 #include <sys/stat.h>
 #include <sys/wait.h>
@@ -35,8 +36,8 @@ static char *_inet_c_ident_ = "@(#)$Id: inet.c,v 1.5 1996/12/22 01:48:54 steve E
 #include <unistd.h>
 #include <dirent.h>
 #include <fcntl.h>
-#include <sys/socket.h>
 #include <netdb.h>
+#include <sys/socket.h>
 #include <netinet/in.h>
 #include <arpa/inet.h>
 #include <time.h>
@@ -47,6 +48,11 @@ static char *_inet_c_ident_ = "@(#)$Id: inet.c,v 1.5 1996/12/22 01:48:54 steve E
 #include "list.h"
 #include "cddbd.h"
 
+#ifndef INADDR_NONE
+/* Solaris does not know INADDR_NONE */
+#define INADDR_NONE     -1
+#endif
+
 /* Structure definitions. */
 
 /* Values for si_flags. */
@@ -76,65 +82,108 @@ int comp_domain(char *, char *);
 int comp_ip(char *, char *);
 int domain_len(char *);
 int isipaddr(char *);
+int read_waitsecs(int, int);
 int sock_getc(void);
 int sock_open(char *, int, int, char *);
 int sock_read(char *, int, int);
 int sock_serv_read(char *, int, int, char *);
 int sock_wait(void);
-int sock_write(char *, int);
+int sock_write(char *, int, int);
+int smtp_write_body(FILE *, int, int, char *);
+int smtp_write_bound(char *, int);
+int smtp_write_hdr(char *);
+int smtp_write_hdr_crc(char *);
+int smtp_write_hdr_enc(int);
+int smtp_write_hdr_type(char *, char *, int);
 
-void cddbd_bcopy(char *, char *, int);
 void chldhand(int);
 void sock_close(void);
+extern void cddbd_translate(char *, char, char);
 
+static char *cmd_send_subj2 = "cddb #command %.16s";    /* originally from mail.c */
 
 /* Global variables. */
 
-int sock_i;
-int sock_e;
-int sock_fd = -1;
+static int sock_i;
+static int sock_e;
+static int sock_fd = -1;
+static int fp_isend = 1;
 
-FILE *sock_fp;
+static FILE *sock_fp;
 
-sinfo_t sinfo;
+static sinfo_t sinfo;
 
-char cddbphost[CDDBHOSTNAMELEN + 1];
+static char cddbphost[CDDBHOSTNAMELEN + 1];
 
 /* Default services table. */
-cservent_t servtab[] = {
-	"cddbp", "tcp", 888,
-	"smtp",  "tcp", 25,
+static cservent_t servtab[] = {
+	{ "cddbp",	"tcp",		8880 },
+	{ "smtp",	"tcp",		25 },
+	{ "http",	"tcp",		80 },
 	{ 0 }
 };
 
 
 void
-get_rmt_hostname(int fd, char *rh)
+get_rmt_hostname(int fd, char *thost, char *rh)
 {
 	char *h;
-	int len;
+	char *a;
+	socklen_t len;
 	struct hostent *hp;
 	struct sockaddr_in name;
 
-	len = sizeof(name);
+	len = (socklen_t)sizeof(name);
 
-	/* Get the peer name. If this call fails, the fd is a tty or file. */
-	if(getpeername(fd, (struct sockaddr *)&name, &len) < 0) {
+	switch(interface) {
+	case IF_HTTP:
+		/* If no remote address, something's very wrong. */
+		if((a = (char *)getenv("REMOTE_ADDR")) == NULL) {
+			sinfo.si_flags |= SF_NOADDR;
+			hp = 0;
+		}
+		else {
+			sinfo.si_haddr.s_addr = inet_addr(a);
+			hp = gethostbyaddr((char *)&sinfo.si_haddr,
+			    sizeof(sinfo.si_haddr), AF_INET);
+		}
+
+		break;
+
+	case IF_CDDBP:
+		/* If we can't get the peername, this is a tty or file. */
+		if(getpeername(fd, (struct sockaddr *)&name, &len) < 0) {
+			sinfo.si_flags |= SF_NOADDR;
+			hp = gethostbyname(host);
+		}
+		else {
+			sinfo.si_haddr = name.sin_addr;
+			hp = gethostbyaddr((char *)&name.sin_addr,
+			    sizeof(name.sin_addr), AF_INET);
+		}
+
+		break;
+
+	case IF_EMAIL:
+	case IF_SUBMIT:
+	case IF_NONE:
+	default:
+		/* In this case, the address may be in the args. */
 		sinfo.si_flags |= SF_NOADDR;
-		hp = gethostbyname(host);
-	}
-	else {
-		sinfo.si_haddr = name.sin_addr;
-		hp = gethostbyaddr((char *)&name.sin_addr,
-		    sizeof(name.sin_addr), AF_INET);
+		hp = 0;
+		break;
 	}
 
 	if(hp == 0) {
-		sinfo.si_flags |= SF_NOHOST;
-		if(sinfo.si_flags & SF_NOADDR)
-			h = "localhost";
-		else
-			h = inet_ntoa(name.sin_addr);
+		if(thost != 0)
+			h = thost;
+		else {
+			sinfo.si_flags |= SF_NOHOST;
+			if(sinfo.si_flags & SF_NOADDR)
+				h = LHOST;
+			else
+				h = inet_ntoa(name.sin_addr);
+		}
 	}
 	else
 		h = (char *)hp->h_name;
@@ -172,7 +221,7 @@ match_host(char *ahost)
 
 	/* Check even if we already have a domain match. */
 	if(sinfo.si_match <= SI_MATCH_DOMAIN && !strncmp("*.", ahost, 2) &&
-	    (sinfo.si_domain_len < domain_len(p)) && !(sinfo.si_match &
+	    (sinfo.si_domain_len <= domain_len(p)) && !(sinfo.si_flags &
 	    SF_NOHOST) && !isipaddr(sinfo.si_hname)) {
 		hp = gethostbyname(sinfo.si_hname);
 
@@ -246,7 +295,7 @@ match_host(char *ahost)
 			sinfo.si_match = SI_MATCH_NET;
 			return 1;
 		}
-
+#ifndef __CYGWIN32__
 		/* Try to match the net name. */
 		np = getnetbyname(ahost);
 
@@ -259,6 +308,7 @@ match_host(char *ahost)
 				return 1;
 			}
 		}
+#endif /* __CYGWIN32__ */
 	}
 
 	return 0;
@@ -304,12 +354,14 @@ comp_domain(char *host, char *domain)
 	int len;
 
 	len = strlen(host) - strlen(domain);
-	if(len <= 0)
+	if(len < 0)
 		return 1;
 
-	host = &host[len];
-	if(*(host - 1) != '.')
-		return 1;
+	if(len > 0) {
+		host = &host[len];
+		if(*(host - 1) != '.')
+			return 1;
+	}
 
 	return(strcmp(host, domain));
 }
@@ -361,15 +413,16 @@ isipaddr(char *ip)
 void
 cddbd_stand(int port)
 {
-	int f;
 	int s;
 	int sock;
 	int opt;
-	int len;
+	socklen_t len;
+	pid_t f;
 	struct sockaddr_in sin;
 	struct sockaddr_in from;
 
-	strcpy(rhost, "none");
+	/* Read the access file. */
+	cddbd_check_access();
 
 	sin.sin_port = get_serv_port(port, SERV_CDDBP);
 
@@ -398,7 +451,7 @@ cddbd_stand(int port)
 
 	signal(SIGCHLD, chldhand);
 
-	f = fork();
+	f = cddbd_fork();
 	if(f < 0) {
 		perror("fork");
 		exit(QUIT_ERR);
@@ -407,16 +460,21 @@ cddbd_stand(int port)
 	if(f > 0)
 		exit(QUIT_OK);
 
+	if(verbose) {
+		printf("Running standalone at port %d.\n", port);
+		fflush(stdout);
+	}
+
 	if(setsid() == -1)
 		fprintf(stderr, "Warning: Can't set process group.\n");
 
-	if(verbose)
-		printf("Running standalone at port %d.", port);
-
-	len = sizeof(from);
+	len = (socklen_t)sizeof(from);
 
 	for(;;) {
+		/* Wait for connections or SIGHUP to reread the access file. */
+		signal(SIGHUP, sighand);
 		sock = accept(s, (struct sockaddr *)&from, &len);
+		signal(SIGHUP, SIG_IGN);
 
 		if(sock < 0) {
 			if(errno == EINTR)
@@ -426,7 +484,7 @@ cddbd_stand(int port)
 			quit(QUIT_ERR);
 		}
 
-		f = fork();
+		f = cddbd_fork();
 
 		if(f < 0) {
 			perror("fork");
@@ -435,7 +493,6 @@ cddbd_stand(int port)
 
 		/* We're the child - crank. */
 		if(f == 0) {
-			curpid = getpid();
 			dup2(sock, 0);
 			close(sock);
 			return;
@@ -447,23 +504,31 @@ cddbd_stand(int port)
 
 
 int
-cddbp_open(char *site, int port)
+cddbp_open(csite_t *sp)
 {
+	int port;
+	ct_key_t *key;
+	uint32_t salt;
+	char *p;
+	char *site;
 	char buf[CDDBBUFSIZ];
 	char errstr[CDDBBUFSIZ];
 
+	site = sp->st_name;
+	port = sp->st_port;
+
 	if(verbose)
-		cddbd_log(LOG_INFO, "Opening CDDBP server: %s", site);
+		cddbd_log(LOG_INFO, "Opening server: %s", site);
 
 	if(!sock_open(site, SERV_CDDBP, port, errstr)) {
 		cddbd_log(LOG_ERR | LOG_NET,
-		    "Failed to open %s CDDBP server: %s.", site, errstr);
+		    "Failed to open %s server: %s.", site, errstr);
 		return 0;
 	}
 
 	if(!sock_read(buf, sizeof(buf), 1)) {
 		cddbd_log(LOG_ERR | LOG_NET,
-		    "Failed to init %s CDDBP server: connection closed.", site);
+		    "Failed to init %s server: connection closed.", site);
 		sock_close();
 		return 0;
 	}
@@ -472,7 +537,7 @@ cddbp_open(char *site, int port)
 	if(buf[0] != '2') {
 		strip_crlf(buf);
 		cddbd_log(LOG_ERR | LOG_NET,
-		    "Failed to init %s CDDBP server: unexpected server "
+		    "Failed to init %s server: unexpected server "
 		    "response: %s.", site, buf);
 
 		sock_close();
@@ -482,9 +547,9 @@ cddbp_open(char *site, int port)
 	cddbd_snprintf(buf, sizeof(buf), "cddb hello %s %s cddbd %sPL%d\n",
 	    user, host, VERSION, PATCHLEVEL);
 
-	if(!sock_write(buf, 1)) {
+	if(!sock_write(buf, 0, 1)) {
 		cddbd_log(LOG_ERR | LOG_NET,
-		    "Failed to handshake with %s CDDBP server (%s).",
+		    "Failed to handshake with %s server (%s).",
 		    site, errno);
 		sock_close();
 		return 0;
@@ -492,7 +557,8 @@ cddbp_open(char *site, int port)
 
 	if(!sock_serv_read(buf, sizeof(buf), 200, errstr)) {
 		cddbd_log(LOG_ERR | LOG_NET,
-		    "Failed to open %s CDDBP server: %s", site, errstr);
+		    "Failed to open %s server: %s", site, errstr);
+
 		sock_close();
 		return 0;
 	}
@@ -500,41 +566,217 @@ cddbp_open(char *site, int port)
 	strncpy(cddbphost, site, sizeof(cddbphost));
 	cddbphost[sizeof(cddbphost) - 1] = '\0';
 
+	/* If we don't need to validate, return now. */
+	if(!(sp->st_flags & ST_FLAG_PWDLBL))
+		return 1;
+
+	if((key = getpasswd(sp->st_pwdlbl)) == NULL) {
+		cddbd_log(LOG_ERR | LOG_NET,
+		    "No password label \"%s\" found for site: %s",
+		    sp->st_pwdlbl, sp->st_name);
+
+		sock_close();
+		return 0;
+	}
+
+	if(!sock_write("validate\n", 0, 1)) {
+		cddbd_log(LOG_ERR | LOG_NET,
+		    "Failed to perform validation with server: %s",
+		    site);
+
+		sock_close();
+		return 0;
+	}
+
+	if(!sock_serv_read(buf, sizeof(buf), 320, errstr)) {
+		cddbd_log(LOG_ERR | LOG_NET,
+		    "Failed to perform validation with server: %s",
+		    site);
+
+		sock_close();
+		return 0;
+	}
+	
+	if((p = strstr(buf, saltstr)) == NULL) {
+		cddbd_log(LOG_ERR | LOG_NET,
+		    "Malformed %s response from %s server: %s", site, buf);
+
+		(void)sock_write("\n", 0, 1);
+		sock_close();
+		return 0;
+	}
+
+	p += strlen(saltstr);
+
+	if(sscanf(p, "%08x", &salt) != 1) {
+		cddbd_log(LOG_ERR | LOG_NET,
+		    "Malformed %s response from %s server: %s", site, buf);
+
+		(void)sock_write("\n", 0, 1);
+		sock_close();
+		return 0;
+	}
+
+	p = crctostr(CT_PASSWD, salt, key);
+
+	cddbd_snprintf(buf, sizeof(buf), "%s\n", p);
+
+	if(!sock_write(buf, 0, 1)) {
+		cddbd_log(LOG_ERR | LOG_NET,
+		    "Failed to perform validation with server: %s",
+		    site);
+
+		sock_close();
+		return 0;
+	}
+
+	if(!sock_serv_read(buf, sizeof(buf), 200, errstr)) {
+		cddbd_log(LOG_ERR | LOG_NET,
+		    "Failed to perform validation with server: %s",
+		    site);
+
+		sock_close();
+		return 0;
+	}
+
 	return 1;
 }
 
 
+int
+cddbp_command(char *command, int lev)
+{
+	char *p;
+	char code[4];
+	char buf[CDDBBUFSIZ];
+	/* char errstr[CDDBBUFSIZ]; */
+
+	cddbd_snprintf(buf, sizeof(buf), "proto %d\n", lev);
+
+	if(!sock_write(buf, 0, 1)) {
+		cddbd_log(LOG_ERR | LOG_NET,
+		    "Failed to set protocol level (%d) on %s server (%d)",
+		    lev, cddbphost, errno);
+		return 0;
+	}
+
+	if(!sock_read(buf, sizeof(buf), 1) ||
+	    (sscanf(buf, "%03s", code) != 1) ||
+	    (strcmp(code, "201") && strcmp(code, "502"))) {
+		strip_crlf(buf);
+		cddbd_log(LOG_ERR | LOG_NET,
+		    "Failed to set protocol level (%d) on %s server: %s",
+		    lev, cddbphost, buf);
+
+		return 0;
+	}
+
+	cddbd_snprintf(buf, sizeof(buf), "%s\n", command);
+
+	if(!sock_write(buf, 0, 1)) {
+		cddbd_log(LOG_ERR | LOG_NET,
+		    "Failed to send command \"%s\" to %s server (%d)",
+		    command, cddbphost, errno);
+		return 0;
+	}
+
+	if(!sock_read(buf, sizeof(buf), 1) ||
+	    (sscanf(buf, "%03s", code) != 1) || code[0] != '2') {
+		strip_crlf(buf);
+		cddbd_log(LOG_ERR | LOG_NET,
+		    "Unexpected response from server %s for command \"%s\": %s",
+		    cddbphost, command, buf);
+
+		return 0;
+	}
+
+	switch(code[1]) {
+	case '0':
+		p = buf;
+
+		while(*p != '\0' && !isspace(*p))
+			p++;
+		while(*p != '\0' && isspace(*p))
+			p++;
+
+		strip_crlf(p);
+		printf("%s\n", p);
+
+		return 1;
+
+	case '1':
+		while(sock_read(buf, sizeof(buf), 0)) {
+			if(is_dot(buf))
+				return 1;
+
+			p = buf;
+
+			if(is_dbl_dot(p))
+				p++;
+
+			strip_crlf(p);
+			printf("%s\n", p);
+		}
+
+		return 0;
+
+	default:
+		cddbd_log(LOG_ERR | LOG_NET,
+		    "Failed to send command \"%s\" to %s server: %s (%d)",
+		    command, cddbphost, code, errno);
+
+		return 0;
+	}
+}
+
+
 int
 cddbp_log_stats(void)
 {
+	char *p;
 	char buf[CDDBBUFSIZ];
 	char errstr[CDDBBUFSIZ];
 
-	if(!sock_write("log day\n", 1)) {
+	/* Try to get a long list first, but not all servers support this. */
+	if(!sock_write("log -l 200 day\n", 0, 1)) {
 		cddbd_log(LOG_ERR | LOG_NET,
-		    "Failed to send log command to %s CDDBP server (%s)",
+		    "Failed to send log command to %s server (%d)",
 		    cddbphost, errno);
 		return 0;
 	}
 
+	/* If it didn't work, try a more generic log command. */
 	if(!sock_serv_read(buf, sizeof(buf), 210, errstr)) {
-		cddbd_log(LOG_ERR | LOG_NET,
-		    "Failed to get log stats from %s CDDBP server: %s",
-		    cddbphost, errstr);
-		return 0;
+		if(!sock_write("log day\n", 0, 1)) {
+			cddbd_log(LOG_ERR | LOG_NET,
+			    "Failed to send log command to %s server (%d)",
+			    cddbphost, errno);
+			return 0;
+		}
+
+		if(!sock_serv_read(buf, sizeof(buf), 210, errstr)) {
+			cddbd_log(LOG_ERR | LOG_NET,
+			    "Failed to get log stats from %s server: %s",
+			    cddbphost, errstr);
+			return 0;
+		}
 	}
 
 	while(sock_read(buf, sizeof(buf), 0)) {
-		strip_crlf(buf);
-
-		if(!strcmp(buf, "."))
+		if(is_dot(buf))
 			return 1;
 
-		printf("%s\n", buf);
+		p = buf;
+
+		if(is_dbl_dot(p))
+			p++;
+
+		strip_crlf(p);
+		printf("%s\n", p);
 	}
 
 	cddbd_log(LOG_ERR | LOG_NET,
-	    "Failed to get log stats from %s CDDBP server: "
+	    "Failed to get log stats from %s server: "
 	    "connection closed (%d)", cddbphost, errno);
 
 	return 0;
@@ -550,32 +792,32 @@ cddbp_transmit(db_t *db, char *category, unsigned int discid)
 	cddbd_snprintf(buf, sizeof(buf), "cddb write %s %08x\n",
 	    category, discid);
 
-	if(!sock_write(buf, 1)) {
+	if(!sock_write(buf, 0, 1)) {
 		cddbd_log(LOG_ERR | LOG_XMIT,
-		    "Failed to send write command to %s CDDBP server (%s).",
+		    "Failed to send write command to %s server (%s).",
 		    cddbphost, errno);
 		return 0;
 	}
 
 	if(!sock_serv_read(buf, sizeof(buf), 320, errstr)) {
 		cddbd_log(LOG_ERR | LOG_XMIT,
-		    "Failed to send write command to %s CDDBP server: %s.",
+		    "Failed to send write command to %s server: %s.",
 		    cddbphost, errstr);
 		return 0;
 	}
 
-	if(!db_write(sock_fp, db)) {
+	if(!db_write(sock_fp, db, MAX_PROTO_LEVEL)) {
 		cddbd_log(LOG_ERR | LOG_XMIT,
-		    "Failed to write DB entry to %s CDDBP server (%d).",
+		    "Failed to write DB entry to %s server (%d).",
 		    cddbphost, errno);
 		return 0;
 	}
 
 	fflush(sock_fp);
 
-	if(!sock_write(".\n", 1)) {
+	if(!sock_write(".\n", 0, 1)) {
 		cddbd_log(LOG_ERR | LOG_XMIT,
-		    "Failed to write DB entry to %s CDDBP server (%d).",
+		    "Failed to write DB entry to %s server (%d).",
 		    cddbphost, errno);
 		return 0;
 	}
@@ -584,7 +826,7 @@ cddbp_transmit(db_t *db, char *category, unsigned int discid)
 
 	if(!sock_serv_read(buf, sizeof(buf), 200, errstr)) {
 		cddbd_log(LOG_ERR | LOG_XMIT,
-		    "Failed to write DB entry to %s CDDBP server: %s.",
+		    "Failed to write DB entry to %s server: %s.",
 		    cddbphost, errstr);
 
 		/* Was the DB entry bad, or was there a server error? */
@@ -598,23 +840,81 @@ cddbp_transmit(db_t *db, char *category, unsigned int discid)
 }
 
 
+int
+cddbp_put(FILE *fp, char *file)
+{
+	char buf[CDDBBUFSIZ];
+	char errstr[CDDBBUFSIZ];
+
+	cddbd_snprintf(buf, sizeof(buf), "put %s\n", file);
+
+	if(!sock_write(buf, 0, 1)) {
+		cddbd_log(LOG_ERR | LOG_XMIT,
+		    "Failed to send put command to %s server (%s).",
+		    cddbphost, errno);
+		return 0;
+	}
+
+	if(!sock_serv_read(buf, sizeof(buf), 320, errstr)) {
+		cddbd_log(LOG_ERR | LOG_XMIT,
+		    "Failed to send put command to %s server: %s",
+		    cddbphost, errstr);
+		return 0;
+	}
+
+	fp_copy(fp, sock_fp);
+	fflush(sock_fp);
+	
+	if(!sock_write(".\n", 0, 1)) {
+		cddbd_log(LOG_ERR | LOG_XMIT,
+		    "Failed to write %s to %s server (%d).",
+		    file, cddbphost, errno);
+		return 0;
+	}
+
+	buf[0] = '\0';
+
+	if(!sock_serv_read(buf, sizeof(buf), 200, errstr)) {
+		cddbd_log(LOG_ERR | LOG_XMIT,
+		    "Failed to write %s to %s server: %s.",
+		    file, cddbphost, errstr);
+
+		/* Was the DB entry bad, or was there a server error? */
+		if(!strncmp(buf, "501", 3))
+			return -1;
+		else
+			return 0;
+	}
+
+	return 1;
+}
+
+
 int
 cddbp_update(void)
 {
+	int err;
+	int code;
 	char buf[CDDBBUFSIZ];
 	char errstr[CDDBBUFSIZ];
 
-	if(!sock_write("update\n", 1)) {
+	if(!sock_write("update\n", 0, 1)) {
 		cddbd_log(LOG_ERR | LOG_XMIT,
-		    "Failed to initiate update on %s CDDBP server (%d).",
+		    "Failed to initiate update on %s server (%d).",
 		    cddbphost, errno);
 		return 0;
 	}
 
 	if(!sock_serv_read(buf, sizeof(buf), 200, errstr)) {
-		cddbd_log(LOG_ERR | LOG_XMIT,
-		    "Failed to initiate update on %s CDDBP server: %s.",
+		/* If we're denied permission, that's okay. */
+		if((sscanf(buf, "%d", &code) != 1) || code != 401)
+			err = LOG_ERR | LOG_XMIT;
+		else
+			err = LOG_INFO;
+
+		cddbd_log(err, "Failed to initiate update on %s server: %s.",
 		    cddbphost, errstr);
+
 		return 0;
 	}
 
@@ -625,7 +925,7 @@ cddbp_update(void)
 void
 cddbp_close(void)
 {
-	(void)sock_write("quit\n", 1);
+	(void)sock_write("quit\n", 0, 1);
 	sock_close();
 }
 
@@ -653,7 +953,7 @@ smtp_open()
 		return 0;
 	}
 
-	if(!sock_write("helo\n", 1)) {
+	if(!sock_write("helo\n", 0, 1)) {
 		cddbd_log(LOG_ERR | LOG_NET,
 		    "Failed to handshake with %s SMTP server (%s)",
 		    smtphost, errno);
@@ -679,7 +979,7 @@ smtp_open()
 	/* Try the helo with an argument to see if that helps. */
 	cddbd_snprintf(buf, sizeof(buf), "helo %s\n", host);
 
-	if(!sock_write(buf, 1)) {
+	if(!sock_write(buf, 0, 1)) {
 		cddbd_log(LOG_ERR | LOG_NET,
 		    "Failed to handshake with %s SMTP server (%s)",
 		    smtphost, errno);
@@ -706,20 +1006,115 @@ smtp_open()
 void
 smtp_close(void)
 {
-	(void)sock_write("quit\n", 1);
+	(void)sock_write("quit\n", 0, 1);
 	sock_close();
 }
 
 
+/* return values: 0..error; 1..ok */
 int
-smtp_transmit(FILE *fp, char *subject, char *rcpt, char *hto, int enc, int len)
+smtp_command(csite_t *sp, char *rcmd, int lev, char *hto, char *mfrom)
 {
+	FILE *fp;
+	char *tcmd;
+	char *tfile;
 	char buf[CDDBBUFSIZ];
+	char buf2[CDDBBUFSIZ];
+	ct_key_t *key;
+
+	tfile = cddbd_mktemp();
+
+	if((fp = fopen(tfile, "w+")) == NULL) {
+		cddbd_log(LOG_ERR | LOG_NET,
+		    "Failed to open command temp file: %s (%d)", tfile, errno);
+
+		cddbd_freetemp(tfile);
+		return 0;
+	}
+
+	if(!asy_encode(rcmd, &tcmd)) {
+		cddbd_log(LOG_ERR | LOG_NET,
+		    "Failed encode command: %s (%d)", rcmd, errno);
+
+		fclose(fp);
+		cddbd_freetemp(tfile);
+
+		return 0;
+	}
+
+	cddbd_translate(tcmd, ' ', '+');
+
+	if(fprintf(fp, "cmd=%s&hello=%s+%s+cddbd+%sPL%d&proto=%d\n",
+	    tcmd, user, host, VERSION, PATCHLEVEL, lev) <= 0) {
+		cddbd_log(LOG_ERR | LOG_NET,
+		    "Failed to open command temp file: %s (%d)", tfile, errno);
+
+		fclose(fp);
+		cddbd_freetemp(tfile);
+
+		return 0;
+	}
+
+	fflush(fp);
+	rewind(fp);
+
+	if(sp->st_flags & ST_FLAG_PWDLBL) {
+		if((key = getpasswd(sp->st_pwdlbl)) == NULL) {
+			cddbd_log(LOG_ERR | LOG_XMIT, "No password "
+			    "label \"%s\" found for site: %s",
+			    sp->st_pwdlbl, sp->st_name);
+
+			fclose(fp);
+			cddbd_freetemp(tfile);
+
+			return 0;
+		}
+	}
+	else
+		key = 0;
+
+	cddbd_snprintf(buf2, sizeof(buf2), "cmd_%08x", cddbd_rand());
+	cddbd_snprintf(buf, sizeof(buf), cmd_send_subj2, buf2);
+
+	if(!smtp_transmit(fp, buf, hto, hto, mfrom, 0, MF_MULTI, 0, key)) {
+		cddbd_log(LOG_ERR | LOG_XMIT,
+		    "Failed to send command \"%s\" to %s server (%d)",
+		    rcmd, smtphost, errno);
+
+		fclose(fp);
+		cddbd_freetemp(tfile);
+
+		return 0;
+	}
+
+	fclose(fp);
+	cddbd_freetemp(tfile);
+	return 1;
+}
+
+
+int
+smtp_transmit(FILE *fp, char *subject, char *rcpt, char *hto, char *mfrom,
+    char *echo, int flags, int len, ct_key_t *key)
+{
+	char *crcstr;
+	char buf[CDDBBUFSIZ];
+	char bound[CDDBBUFSIZ];
 	char errstr[CDDBBUFSIZ];
+	uint32_t crc;
 
-	cddbd_snprintf(buf, sizeof(buf), "mail from: %s\n", admin_email);
+	cddbd_snprintf(buf, sizeof(buf), "mail from: %s\n", mfrom);
 
-	if(!sock_write(buf, 1)) {
+	if(crc_email(&crc, fp, admin_email, subject, errstr) != EE_OK) {
+		cddbd_log(LOG_ERR | LOG_NET,
+		    "Failed to generate email crc: %s", errstr);
+
+		return 0;
+	}
+
+	crcstr = crctostr(crc, 0, key);
+
+	if(!sock_write(buf, 0, 1)) {
 		cddbd_log(LOG_ERR | LOG_NET,
 		    "Failed to specify return addr to %s SMTP server (%d)",
 		    smtphost, errno);
@@ -735,7 +1130,7 @@ smtp_transmit(FILE *fp, char *subject, char *rcpt, char *hto, int enc, int len)
 
 	cddbd_snprintf(buf, sizeof(buf), "rcpt to: %s\n", rcpt);
 
-	if(!sock_write(buf, 1)) {
+	if(!sock_write(buf, 0, 1)) {
 		cddbd_log(LOG_ERR | LOG_NET,
 		    "Failed to specify recipient to %s SMTP server (%d)",
 		    smtphost, errno);
@@ -751,7 +1146,7 @@ smtp_transmit(FILE *fp, char *subject, char *rcpt, char *hto, int enc, int len)
 
 	strcpy(buf, "data\n");
 
-	if(!sock_write(buf, 1)) {
+	if(!sock_write(buf, 0, 1)) {
 		cddbd_log(LOG_ERR | LOG_NET,
 		    "Failed to send data to %s SMTP server (%d)",
 		    smtphost, errno);
@@ -765,108 +1160,101 @@ smtp_transmit(FILE *fp, char *subject, char *rcpt, char *hto, int enc, int len)
 		return 0;
 	}
 
-	cddbd_snprintf(buf, sizeof(buf), "%s: %s\n", rpath, admin_email);
+	cddbd_snprintf(buf, sizeof(buf), "%s: %s\n", rpath, mfrom);
 
-	if(!sock_write(buf, 0)) {
-		cddbd_log(LOG_ERR | LOG_NET,
-		    "Failed to write email header to %s SMTP server (%d)",
-		    smtphost, errno);
+	if(!smtp_write_hdr(buf))
 		return 0;
+
+	cddbd_snprintf(buf, sizeof(buf), "%s: %s\n", from, mfrom);
+
+	if(!smtp_write_hdr(buf))
+		return 0;
+
+	cddbd_snprintf(buf, sizeof(buf), "%s: %s\n", x_cddbd_from, admin_email);
+
+	if(!smtp_write_hdr(buf))
+		return 0;
+
+	if(echo != 0) {
+		cddbd_snprintf(buf, sizeof(buf), "%s: %s\n", x_cddbd_echo,
+		    echo);
+
+		if(!smtp_write_hdr(buf))
+			return 0;
 	}
 
 	cddbd_snprintf(buf, sizeof(buf), "%s: CDDBD v%sPL%d\n",
-	    xsender, VERSION, PATCHLEVEL);
+	    x_sender, VERSION, PATCHLEVEL);
 
-	if(!sock_write(buf, 0)) {
-		cddbd_log(LOG_ERR | LOG_NET,
-		    "Failed to write email header to %s SMTP server (%d)",
-		    smtphost, errno);
+	if(!smtp_write_hdr(buf))
 		return 0;
-	}
 
 	cddbd_snprintf(buf, sizeof(buf), "%s: %s\n", to, hto);
 
-	if(!sock_write(buf, 0)) {
-		cddbd_log(LOG_ERR | LOG_NET,
-		    "Failed to write email header to %s SMTP server (%d)",
-		    smtphost, errno);
+	if(!smtp_write_hdr(buf))
 		return 0;
-	}
 
 	cddbd_snprintf(buf, sizeof(buf), "%s: %s\n", subj, subject);
 
-	if(!sock_write(buf, 0)) {
-		cddbd_log(LOG_ERR | LOG_NET,
-		    "Failed to write email header to %s SMTP server (%d)",
-		    smtphost, errno);
+	if(!smtp_write_hdr(buf))
 		return 0;
-	}
 
 	cddbd_snprintf(buf, sizeof(buf), "%s: %s\n", mime_ver, "1.0");
 
-	if(!sock_write(buf, 0)) {
-		cddbd_log(LOG_ERR | LOG_NET,
-		    "Failed to write email header to %s SMTP server (%d)",
-		    smtphost, errno);
+	if(!smtp_write_hdr(buf))
 		return 0;
-	}
 
-	cddbd_snprintf(buf, sizeof(buf), "%s: text/plain; charset=%s\n",
-	    content_type, char_types[CC_ISO_8859_1]);
+	if(flags & MF_MULTI) {
+		cddbd_snprintf(bound, sizeof(bound), "---=====cddbd_%08X%08X_",
+		    cddbd_rand(), cddbd_rand());
 
-	if(!sock_write(buf, 0)) {
-		cddbd_log(LOG_ERR | LOG_NET,
-		    "Failed to write email header to %s SMTP server (%d)",
-		    smtphost, errno);
-		return 0;
-	}
-
-	cddbd_snprintf(buf, sizeof(buf), "%s: %s\n", content_encoding,
-	    encoding_types[enc].en_type);
-
-	if(!sock_write(buf, 0)) {
-		cddbd_log(LOG_ERR | LOG_NET,
-		    "Failed to write email header to %s SMTP server (%d)",
-		    smtphost, errno);
-		return 0;
-	}
-
-	/* If we have a length, and we're not going to change it, print it. */
-	if(len > 0 && encoding_types[enc].en_mapfunc == 0) {
-		cddbd_snprintf(buf, sizeof(buf), "%s: %d\n", content_len, len);
-
-		if(!sock_write(buf, 0)) {
-			cddbd_log(LOG_ERR | LOG_NET,
-			    "Failed to write email header to %s SMTP "
-			    "server (%d)", smtphost, errno);
+		if(!smtp_write_hdr_type(multi_alt, bound, CC_ISO_8859_1))
 			return 0;
-		}
-	}
 
-	/* Separate the header from the body. */
-	if(!sock_write("\n", 0)) {
-		cddbd_log(LOG_ERR | LOG_NET,
-		    "Failed to write email header to %s SMTP server (%d)",
-		    smtphost, errno);
-		return 0;
-	}
-
-	while(fgets(buf, sizeof(buf), fp) != NULL) {
-		/* Remap the charset if needed. */
-		if(encoding_types[enc].en_mapfunc != 0)
-			(encoding_types[enc].en_mapfunc)(buf);
-
-		if(!sock_write(buf, 0)) {
+		/* Separate the header from the body. */
+		if(!sock_write("\n", 0, 0)) {
 			cddbd_log(LOG_ERR | LOG_NET,
-			    "Failed to write mail to %s SMTP server (%d)",
+			    "Failed to write email header to %s SMTP server (%d)",
 			    smtphost, errno);
 			return 0;
 		}
+
+		if(!smtp_write_bound(bound, 0))
+			return 0;
+
+		if(!smtp_write_body(fp, CE_BASE64, CC_ISO_8859_1, crcstr))
+			return 0;
+
+		if(!smtp_write_bound(bound, 0))
+			return 0;
+
+		if(!smtp_write_body(fp, CE_QUOTED_PRINT, CC_ISO_8859_1, crcstr))
+			return 0;
+
+		if(!smtp_write_bound(bound, 1))
+			return 0;
+	}
+	else if(flags & MF_ENC) {
+		if(!smtp_write_body(fp, CE_QUOTED_PRINT, CC_ISO_8859_1, crcstr))
+			return 0;
+	}
+	else {
+		/* If we have a length, print it. */
+		if(len > 0) {
+			cddbd_snprintf(buf, sizeof(buf), "%s: %d\n",
+			    content_len, len);
+
+			if(!smtp_write_hdr(buf))
+				return 0;
+		}
+
+		if(!smtp_write_body(fp, CE_7BIT, CC_US_ASCII, crcstr))
+			return 0;
 	}
 
-	if(!sock_write(".\n", 0)) {
+	if(!sock_write(".\n", 0, 0)) {
 		cddbd_log(LOG_ERR | LOG_NET,
-		    "Failed to send data %s SMTP server (%d)",
+		    "Failed to send data to %s SMTP server (%d)",
 		    smtphost, errno);
 		return 0;
 	}
@@ -882,6 +1270,260 @@ smtp_transmit(FILE *fp, char *subject, char *rcpt, char *hto, int enc, int len)
 }
 
 
+int
+smtp_write_body(FILE *fp, int enc, int set, char *str)
+{
+	int op;
+	int len;
+	char *p;
+	char *p2;
+	char buf[CDDBBUFSIZ];
+
+	/* in some cases, the file needs to be encoded */
+	int encflg= 0;
+	char *tfile= (char *) 0;
+	FILE *tfp= (FILE *) 0;
+	
+	if(!smtp_write_hdr_type(text_plain, 0, set))
+		return 0;
+
+	if(!smtp_write_hdr_enc(enc))
+		return 0;
+
+	if(!smtp_write_hdr_crc(str))
+		return 0;
+
+	/* Separate the header from the body. */
+	if(!sock_write("\n", 0, 0)) {
+		cddbd_log(LOG_ERR | LOG_NET,
+		    "Failed to write email header to %s SMTP server (%d)",
+		    smtphost, errno);
+		return 0;
+	}
+
+	/* Initiate encoding. */
+	if(encoding_types[enc].en_encode != 0) {
+		tfile = cddbd_mktemp();
+
+		if((tfp = fopen(tfile, "w+")) == NULL) {
+			cddbd_log(LOG_ERR | LOG_NET,
+			    "Can't open SMTP tmp file %s (%d)", tfile, errno);
+
+			cddbd_freetemp(tfile);
+			return 0;
+		}
+
+		if((encoding_types[enc].en_encode)(EO_START, 0, 0, 0)
+		    != EN_OK) {
+			cddbd_log(LOG_ERR | LOG_NET,
+			    "Failed to initiate email encoding.");
+
+			cddbd_freetemp(tfile);
+			fclose(tfp);
+
+			return 0;
+		}
+
+		rewind(fp);
+
+		do {
+			p2 = fgets(buf, sizeof(buf), fp);
+
+			if(p2 == NULL)
+				op = EO_END;
+			else {
+				op = EO_ENCODE;
+				len = strlen(buf);
+			}
+
+			if((encoding_types[enc].en_encode)(op,
+			    (unsigned char *)buf, (unsigned char **)&p, &len)
+			    != EN_OK) {
+				cddbd_log(LOG_ERR | LOG_NET,
+				    "Failed to perform email encoding.");
+
+				cddbd_freetemp(tfile);
+				fclose(tfp);
+
+				return 0;
+			}
+
+			if(len > 0 && fwrite(p, 1, len, tfp) != len) {
+				cddbd_log(LOG_ERR | LOG_NET,
+				    "Can't write SMTP tmp file %s (%d)",
+				    tfile, errno);
+
+				cddbd_freetemp(tfile);
+				fclose(tfp);
+
+				return 0;
+			}
+		} while(p2 != NULL);
+
+		encflg = 1;
+		fflush(tfp);
+		fp = tfp;
+	}
+
+	rewind(fp);
+
+	while(fgets(buf, sizeof(buf), fp) != NULL) {
+		if(read_waitsecs(sock_fd, 0) > 0) {
+			if(!sock_read(buf, sizeof(buf), 0)) {
+				cddbd_log(LOG_ERR | LOG_NET,
+				    "Premature output from %s SMTP server.",
+				    smtphost);
+			}
+			else {
+				cddbd_log(LOG_ERR | LOG_NET,
+				    "Premature output from %s SMTP server: %s",
+				    smtphost, buf);
+			}
+
+			if(encflg) {
+				cddbd_freetemp(tfile);
+				fclose(tfp);
+			}
+
+			return 0;
+		}
+
+		len = strlen(buf);
+
+		/* Escape standalone dots in the mail body. */
+		if(is_dbl_dot(buf) && !sock_write(".", 0, 0)) {
+			cddbd_log(LOG_ERR | LOG_NET,
+			    "Failed to write mail to %s SMTP server "
+			    "(%d)", smtphost, errno);
+
+			if(encflg) {
+				cddbd_freetemp(tfile);
+				fclose(tfp);
+			}
+
+			return 0;
+		}
+
+		if(!sock_write(buf, 0, 0)) {
+			cddbd_log(LOG_ERR | LOG_NET,
+			    "Failed to write mail to %s SMTP server (%d)",
+			    smtphost, errno);
+
+			if(encflg) {
+				cddbd_freetemp(tfile);
+				fclose(tfp);
+			}
+
+			return 0;
+		}
+	}
+
+	/* Make sure the last line ends with a newline. */
+	if(buf[strlen(buf) - 1] != '\n' && !sock_write("\n", 0, 0)) {
+		cddbd_log(LOG_ERR | LOG_NET,
+		    "Failed to write mail to %s SMTP server (%d)",
+		    smtphost, errno);
+
+		if(encflg) {
+			cddbd_freetemp(tfile);
+			fclose(tfp);
+		}
+
+		return 0;
+	}
+
+	if(encflg) {
+		cddbd_freetemp(tfile);
+		fclose(tfp);
+	}
+
+	return 1;
+}
+
+
+int
+smtp_write_hdr(char *hdr)
+{
+	if(!sock_write(hdr, 0, 0)) {
+		cddbd_log(LOG_ERR | LOG_NET,
+		    "Failed to write email header to %s SMTP server (%d)",
+		    smtphost, errno);
+		return 0;
+	}
+
+	return 1;
+}
+
+
+int
+smtp_write_hdr_crc(char *str)
+{
+	char buf[CDDBBUFSIZ];
+
+	cddbd_snprintf(buf, sizeof(buf), "%s: %s\n", x_cddbd_crc, str);
+
+	return(smtp_write_hdr(buf));
+}
+
+
+int
+smtp_write_hdr_enc(int enc)
+{
+	char buf[CDDBBUFSIZ];
+
+	cddbd_snprintf(buf, sizeof(buf), "%s: %s\n", content_encoding,
+	    encoding_types[enc].en_type);
+
+	return(smtp_write_hdr(buf));
+}
+
+
+int
+smtp_write_bound(char *bound, int end)
+{
+	char buf[CDDBBUFSIZ];
+
+	if(end)
+		cddbd_snprintf(buf, sizeof(buf), "--%s--\n", bound);
+	else
+		cddbd_snprintf(buf, sizeof(buf), "--%s\n", bound);
+
+	if(!sock_write(buf, 0, 0)) {
+		cddbd_log(LOG_ERR | LOG_NET,
+		    "Failed to write mail boundary to %s SMTP server (%d)",
+		    smtphost, errno);
+
+		return 0;
+	}
+
+	return 1;
+}
+
+
+int
+smtp_write_hdr_type(char *type, char *bound, int set)
+{
+	char buf[CDDBBUFSIZ];
+
+	if(bound != 0) {
+		cddbd_snprintf(buf, sizeof(buf), "%s: %s; %s=%s;\n",
+		    content_type, type, charset, char_types[set]);
+
+		if(!smtp_write_hdr(buf))
+			return 0;
+
+		cddbd_snprintf(buf, sizeof(buf), "\t%s=\"%s\"\n",
+		    boundary, bound);
+	}
+	else {
+		cddbd_snprintf(buf, sizeof(buf), "%s: %s; %s=%s\n",
+		    content_type, type, charset, char_types[set]);
+	}
+
+	return(smtp_write_hdr(buf));
+}
+
+
 int
 sock_open(char *shost, int ind, int rport, char *errstr)
 {
@@ -890,8 +1532,8 @@ sock_open(char *shost, int ind, int rport, char *errstr)
 	struct hostent *hp;
 	struct sockaddr_in sin;
 
-	if(sock_fd != -1)
-		return 0;
+	if(sock_fd >= 0)
+		sock_close();
 
 	port = get_serv_port(rport, ind);
 
@@ -899,7 +1541,10 @@ sock_open(char *shost, int ind, int rport, char *errstr)
 		cddbd_bcopy(hp->h_addr, (char *)&sin.sin_addr, hp->h_length);
 	}
 	else {
-		if(isipaddr(shost) && (ad.s_addr = inet_addr(shost)) != -1) {
+		/* T2D: inet_addr() should be replaced by inet_aton() which
+                 *      allows a cleaner error handling.
+                 */
+		if(isipaddr(shost) && (ad.s_addr = inet_addr(shost)) != INADDR_NONE) {
 			cddbd_bcopy((char *)&ad.s_addr, (char *)&sin.sin_addr,
 			    sizeof(ad.s_addr));
 		}
@@ -940,7 +1585,7 @@ sock_open(char *shost, int ind, int rport, char *errstr)
 void
 sock_close(void)
 {
-	if(sock_fd == -1)
+	if(sock_fd < 0)
 		return;
 
 	shutdown(sock_fd, 2);
@@ -980,27 +1625,28 @@ sock_serv_read(char *buf, int size, int expect, char *errstr)
 
 
 int
-sock_write(char *buf, int flag)
+sock_write(char *buf, int cnt, int flag)
 {
 	int cc;
-	int cnt;
 	char *p;
 	char *pbuf;
 
+	if(cnt == 0)
+		cnt = strlen(buf);
+
 	p = buf;
-	cnt = strlen(buf);
 
 	while(cnt > 0) {
-		cc = write(sock_fd, buf, cnt);
+		cc = write(sock_fd, p, cnt);
 		if(cc < 0)
 			return 0;
 
-		buf += cc;
+		p += cc;
 		cnt -= cc;
 	}
 
 	if(verbose && flag) {
-		pbuf = strdup(p);
+		pbuf = strdup(buf);
 
 		if(pbuf == NULL)
 			cddbd_log(LOG_ERR, "Can't malloc pbuf.");
@@ -1056,7 +1702,7 @@ sock_read(char *buf, int cnt, int flag)
 int
 sock_getc(void)
 {
-	static char buf[CDDBBUFSIZ];
+	static unsigned char buf[CDDBBUFSIZ];
 
 	if(sock_i == sock_e) {
 		if(!sock_wait())
@@ -1079,32 +1725,23 @@ sock_getc(void)
 
 int
 sock_wait(void)
+{
+	return read_wait(sock_fd);
+}
+
+
+int
+read_wait(int fd)
 {
 	int n;
-	fd_set readfds;
-	struct timeval tv;
 
 	if(xmit_time == 0)
 		return 1;
 
 	do {
-		FD_ZERO(&readfds);
-		FD_SET(sock_fd, &readfds);
-
-		tv.tv_sec = xmit_time;
-		tv.tv_usec = 0;
-
-#ifdef __hpux
-		n = select((sock_fd + 1), (int *)&readfds, (int *)NULL,
-		    (int *)NULL, &tv);
-#else
-		n = select((sock_fd + 1), &readfds, (fd_set *)NULL,
-		    (fd_set *)NULL, &tv);
-#endif
-
+		n = read_waitsecs(fd, xmit_time);
 		if(n > 0)
 			return 1;
-
 	} while(n < 0 && errno == EINTR);
 
 	if(n == 0) {
@@ -1117,13 +1754,90 @@ sock_wait(void)
 }
 
 
+int
+read_waitsecs(int fd, int secs)
+{
+	int n;
+	fd_set readfds;
+	struct timeval tv;
+
+	FD_ZERO(&readfds);
+	FD_SET(fd, &readfds);
+
+	tv.tv_sec = secs;
+	tv.tv_usec = 0;
+
+	n = select((fd + 1), &readfds, (fd_set *)NULL, (fd_set *)NULL, &tv);
+
+	return n;
+}
+
+
 void
 cddbd_bcopy(char *from, char *to, int size)
+{
+	while(--size >= 0)
+		to[size] = from[size];
+}
+
+
+void
+cddbd_bzero(char *addr, int size)
+{
+	while(--size >= 0)
+		addr[size] = 0;
+}
+
+
+void
+fp_copy(FILE *ifp, FILE *ofp)
+{
+	char buf[CDDBBUFSIZ];
+
+	fp_copy_init();
+
+	/* Copy until input is exhausted. */
+	while(fgets(buf, sizeof(buf), ifp))
+		if(fp_copy_buf(ofp, buf) != 0)
+			break;
+}
+
+
+void
+fp_copy_init(void)
+{
+	fp_isend = 1;
+}
+
+
+int
+fp_copy_buf(FILE *ofp, char *p)
+{
+	if((fp_isend && is_dbl_dot(p) && fputc('.', ofp) == EOF) ||
+	    fputs(p, ofp) == EOF) {
+		fp_isend = 1;
+		return 1;
+	}
+
+	if(p[strlen(p) - 1] == '\n')
+		fp_isend = 1;
+	else
+		fp_isend = 0;
+
+	return 0;
+}
+
+
+int
+get_serv_index(char *sname)
 {
 	int i;
 
-	for(i = 0; i < size; i++)
-		to[i] = from[i];
+	for(i = 0; servtab[i].s_name != 0; i++)
+		if(!cddbd_strcasecmp(servtab[i].s_name, sname))
+			return i;
+
+	return -1;
 }
 
 
