@@ -1,10 +1,12 @@
 /*
  *   cddbd - CD Database Protocol Server
  *
- *   Copyright (C) 1996  Steve Scherf
+ *   Copyright (C) 1996-1997  Steve Scherf
  *   Email: steve@moonsoft.com
  *   Moondog Software Productions - makers of fine public domain software.
  *
+ *   07.03.2000 modified to comply protocol version 4 -OGF-
+ *
  *   This program is free software; you can redistribute it and/or modify
  *   it under the terms of the GNU General Public License as published by
  *   the Free Software Foundation; either version 2 of the License, or
@@ -22,8 +24,8 @@
  */
 
 #ifndef LINT
-static char *_patchlevel_h_ident_ = "@(#)$Id: patchlevel.h,v 1.5 1996/12/13 23:58:06 steve Exp $";
+static char *_patchlevel_h_ident_ = "@(#)$Id: patchlevel.h,v 1.9 2001/09/10 15:33:41 joerg78 Exp $";
 #endif
 
-#define VERSION		"1.3.1"		/* The version number. */
+#define VERSION		"1.4"	/* The version number. */
 #define PATCHLEVEL	0		/* The patch level. */
