@@ -1,7 +1,7 @@
 /*
  *   cddbd - CD Database Protocol Server
  *
- *   Copyright (C) 1996  Steve Scherf
+ *   Copyright (C) 1996-1997  Steve Scherf
  *   Email: steve@moonsoft.com
  *   Moondog Software Productions - makers of fine public domain software.
  *
@@ -22,9 +22,11 @@
  */
 
 #ifndef LINT
-static char *_sites_c_ident_ = "@(#)$Id: xmit.c,v 1.1 1996/12/13 23:35:22 steve Exp $";
+static char *_sites_c_ident_ = "@(#)$Id: xmit.c,v 1.8 2001/09/08 18:16:54 gonter Exp $";
 #endif
 
+#include <stdlib.h>
+#include <sys/wait.h>
 #include <sys/types.h>
 #include <sys/stat.h>
 #include <string.h>
@@ -33,16 +35,20 @@ static char *_sites_c_ident_ = "@(#)$Id: xmit.c,v 1.1 1996/12/13 23:35:22 steve
 #include <errno.h>
 #include <dirent.h>
 #include <time.h>
+#include <unistd.h>
+#include <netinet/in.h>
+#include "access.h"
 #include "list.h"
 #include "cddbd.h"
 
 
 /* Preprocessor definitions. */
 
-#define PROT_CDDBP	0
-#define PROT_SMTP	1
-#define PROT_INFO	2
-#define PROT_LOG	3
+/* Sites file record type. */
+#define SRT_CDDBP	0
+#define SRT_SMTP	1
+#define SRT_INFO	2
+#define SRT_GROUP	3
 
 
 /* Structures. */
@@ -57,11 +63,18 @@ typedef struct tfile {
 
 void cddbd_catchup(char *, char *);
 void cddbd_do_catchup(char *, char *);
-void cddbd_do_remote_log(char *, int);
-void cddbd_do_transmit(site_t *);
+void cddbd_do_put(csite_t *, char *);
+void cddbd_do_remote_cmd(csite_t *, char *, char *);
+void cddbd_do_remote_log(csite_t *);
+void cddbd_do_transmit(csite_t *);
+void cddbd_put(char *, char *);
+void cddbd_remote_cmd(char *, char *, char *);
 void cddbd_remote_log(char *);
 void cddbd_transmit(char *);
 void cddbd_write_history(char *, unsigned int);
+void getsitemark(void);
+void gfree(void *);
+void setsitemark(void);
 void tfree(void *);
 
 int cddbd_clean_history(void);
@@ -69,8 +82,10 @@ int cddbd_do_merge_ulist(char *, int);
 int cddbd_oc_history(char *, int);
 int cddbd_open_history(char *);
 int cddbd_open_ulist(void);
+int cddbd_put_file(char *, char *, char *);
 int cddbd_read_history(char *, unsigned int *);
 int parse_coord(char *, coord_t *);
+int grp_comp(void *, void *);
 int tfile_comp(void *, void *);
 
 
@@ -80,18 +95,23 @@ FILE *site_fp = NULL;
 FILE *ulist_fp = NULL;
 
 char *histstr = "Start %s\n";
+char *uhist;
 
 int hist_opened;
 int site_line;
 
+long sitemark = -1;
+
 lhead_t *hist;
+lhead_t *group;
 
 
-site_t *
+csite_t *
 getsitenam(char *site, int type)
 {
-	site_t *sp;
+	csite_t *sp;
 
+	endsitegrp();
 	setsiteent();
 
 	while((sp = getsiteent(type)) != NULL)
@@ -102,14 +122,16 @@ getsitenam(char *site, int type)
 }
 
 
-site_t *
+csite_t *
 getsiteent(int type)
 {
 	int i;
-	int proto;
+	int ind;
+	int srt;
 	arg_t args;
+	char *s;
 	char buf[CDDBBUFSIZ];
-	static struct site site;
+	static csite_t site;
 
 	if(site_fp == NULL) {
 		setsiteent();
@@ -134,31 +156,37 @@ getsiteent(int type)
 			continue;
 		}
 
+		cddbd_bzero((char *)&site, sizeof(site));
 		strncpy(site.st_name, args.arg[0], sizeof(site.st_name));
 		site.st_name[sizeof(site.st_name) - 1] = '\0';
 		site.st_port = 0;
 
 		if(!cddbd_strcasecmp(args.arg[1], "cddbp"))
-			proto = PROT_CDDBP;
+			srt = SRT_CDDBP;
 		else if(!cddbd_strcasecmp(args.arg[1], "smtp"))
-			proto = PROT_SMTP;
-		else if(!cddbd_strcasecmp(args.arg[1], "log"))
-			proto = PROT_LOG;
+			srt = SRT_SMTP;
 		else if(!cddbd_strcasecmp(args.arg[1], "info"))
-			proto = PROT_INFO;
+			srt = SRT_INFO;
+		else if(!cddbd_strcasecmp(args.arg[1], "group"))
+			srt = SRT_GROUP;
 		else {
 			cddbd_log(LOG_ERR,
-			    "Unknown protocol \"%s\" in %s on line %d.",
+			    "Unknown site record type \"%s\" in %s on line %d.",
 			    args.arg[1], sitefile, site_line);
 
 			continue;
 		}
 
+		/* If we have a group filter, make sure the site is included. */
+		if(type != SITE_GROUP && group != 0 &&
+		    list_find(group, (void *)site.st_name) == 0)
+			continue;
+
 		switch(type) {
 		case SITE_XMIT:
-			if(proto == PROT_CDDBP)
+			if(srt == SRT_CDDBP)
 				site.st_proto = ST_PROTO_CDDBP;
-			else if(proto == PROT_SMTP)
+			else if(srt == SRT_SMTP)
 				site.st_proto = ST_PROTO_SMTP;
 			else
 				continue;
@@ -166,7 +194,7 @@ getsiteent(int type)
 			break;
 
 		case SITE_LOG:
-			if(proto == PROT_LOG)
+			if(srt == SRT_CDDBP)
 				site.st_proto = ST_PROTO_CDDBP;
 			else
 				continue;
@@ -174,13 +202,21 @@ getsiteent(int type)
 			break;
 
 		case SITE_INFO:
-			if(proto == PROT_INFO)
+			if(srt == SRT_INFO)
 				site.st_proto = ST_PROTO_NONE;
 			else
 				continue;
 
 			break;
 
+		case SITE_GROUP:
+			if(srt == SRT_GROUP)
+				site.st_proto = ST_PROTO_GROUP;
+			else
+				continue;
+
+			break;
+
 		default:
 			cddbd_log(LOG_ERR,
 			    "Internal error: unknown site type: %s.", type);
@@ -190,54 +226,159 @@ getsiteent(int type)
 
 		switch(site.st_proto) {
 		case ST_PROTO_CDDBP:
-			if(args.nargs > 3) {
+		case ST_PROTO_SMTP:
+			if(args.nargs > 5) {
 				cddbd_log(LOG_ERR,
 				    "Syntax error in %s on line %d.",
 				    sitefile, site_line);
 				continue;
 			}
 
-			if(args.nargs == 3) {
-				site.st_port = atoi(args.arg[2]);
+			site.st_port = 0;
+			site.st_flags = 0;
 
-				if(site.st_port <= 0) {
+			/* Get the port #. */
+			switch(site.st_proto) {
+			case ST_PROTO_CDDBP:
+				if(args.nargs > 2) {
+					if(strcmp(args.arg[2], "-")) {
+						site.st_port =
+						    atoi(args.arg[2]);
+					}
+
+					if(site.st_port < 0) {
+						cddbd_log(LOG_ERR,
+						    "Invalid port in %s on "
+						    "line %d.", sitefile,
+						    site_line);
+
+						continue;
+					}
+				}
+
+				break;
+
+			case ST_PROTO_SMTP:
+				if(args.nargs < 3) {
 					cddbd_log(LOG_ERR,
-					    "Invalid port in %s on line %d.",
+					    "Missing address in %s on line %d.",
 					    sitefile, site_line);
 
 					continue;
 				}
+
+				strncpy(site.st_addr, args.arg[2],
+				    sizeof(site.st_addr));
+
+				break;
+
+			default:
+				break;
+			}
+
+			/* Get the flags. */
+			if(args.nargs > 3 && strcmp(args.arg[3], "-")) {
+				for(s = args.arg[3]; *s != '\0'; s++) {
+					switch(*s) {
+					case 't':
+						site.st_flags |= ST_FLAG_NOXMIT;
+						break;
+
+					case 'm':
+						site.st_flags |= ST_FLAG_OMIME;
+						break;
+
+					default:
+						break;
+					}
+				}
+
+				if(*s != '\0') {
+					cddbd_log(LOG_ERR,
+					    "Invalid flag '%c' in %s on line "
+					    "%d.", *s, sitefile, site_line);
+
+					continue;
+				}
+			}
+
+			/* Get the password label. */
+			if(args.nargs > 4) {
+				site.st_flags |= ST_FLAG_PWDLBL;
+				cddbd_snprintf(site.st_pwdlbl,
+				    sizeof(site.st_pwdlbl), "%s", args.arg[4]);
 			}
-			else
-				site.st_port = 0;
 
 			break;
 
-		case ST_PROTO_SMTP:
-			if(args.nargs != 3) {
+		case ST_PROTO_GROUP:
+			if(args.nargs < 3) {
 				cddbd_log(LOG_ERR,
 				    "Syntax error in %s on line %d.",
 				    sitefile, site_line);
 				continue;
 			}
 
-			site.st_proto = ST_PROTO_SMTP;
+			if(group == 0) {
+				cddbd_log(LOG_ERR, "Internal error: no group"
+				    "list in group search.");
+				quit(1);
+			}
 
-			strncpy(site.st_user, args.arg[2],
-			    sizeof(site.st_user));
+			/* This isn't the group we're looking for. */
+			if(strcmp((char *)group->lh_data, site.st_name))
+				continue;
+
+			for(i = 2; i < args.nargs; i++) {
+				if(list_find(group, (void *)args.arg[i]) != 0) {
+					cddbd_log(LOG_ERR, "Site \"%s\" "
+					    "multiply defined in group \"%s\""
+					    " in %s on line %d.", args.arg[i],
+					    (void *)group->lh_data, sitefile,
+					    site_line);
+					quit(1);
+				}
+
+				s = strdup(args.arg[i]);
+				if(s == NULL) {
+					cddbd_log(LOG_ERR, "Failed to allocate"
+					    "group list string.");
+					quit(1);
+				}
+
+				if(list_add_cur(group, (void *)s) == 0) {
+					cddbd_log(LOG_ERR, "Failed to allocate"
+					    "group list entry.");
+					quit(1);
+				}
+			}
 
 			break;
 
 		case ST_PROTO_NONE:
-			if(args.nargs < 6) {
+			if(args.nargs < 8) {
 				cddbd_log(LOG_ERR,
 				    "Syntax error in %s on line %d.",
 				    sitefile, site_line);
 				continue;
 			}
 
-			if(cddbd_strcasecmp(args.arg[2], "-")) {
-				site.st_port = atoi(args.arg[2]);
+			if((ind = get_serv_index(args.arg[2])) < 0) {
+				cddbd_log(LOG_ERR,
+				    "Invalid service %s in %s on line %d.",
+				    args.arg[2], sitefile, site_line);
+
+				continue;
+			}
+
+			site.st_proto = ind;
+
+			strncpy(site.st_pname, args.arg[2],
+			    (sizeof(site.st_pname) - 1));
+			site.st_pname[strlen(site.st_pname)] = '\0';
+
+			if(cddbd_strcasecmp(args.arg[3], "-")) {
+				site.st_port = atoi(args.arg[3]);
 
 				if(site.st_port <= 0) {
 					cddbd_log(LOG_ERR,
@@ -248,10 +389,13 @@ getsiteent(int type)
 				}
 			}
 			else
-				site.st_port = ntohs(get_serv_port(0,
-				    SERV_CDDBP));
+				site.st_port = ntohs(get_serv_port(0, ind));
 
-			strncpy(buf, args.arg[3], (sizeof(buf) - 1));
+			strncpy(site.st_addr, args.arg[4],
+			    (sizeof(site.st_addr) - 1));
+			site.st_addr[strlen(site.st_addr)] = '\0';
+
+			strncpy(buf, args.arg[5], (sizeof(buf) - 1));
 			buf[strlen(buf)] = '\0';
 
 			if(!parse_coord(buf, &site.st_lat)) {
@@ -261,7 +405,7 @@ getsiteent(int type)
 				continue;
 			}
 
-			strncpy(buf, args.arg[4], (sizeof(buf) - 1));
+			strncpy(buf, args.arg[6], (sizeof(buf) - 1));
 			buf[strlen(buf)] = '\0';
 
 			if(!parse_coord(buf, &site.st_long)) {
@@ -273,7 +417,7 @@ getsiteent(int type)
 
 			site.st_desc[0] = '\0';
 
-			for(i = 5; i < args.nargs; i++) {
+			for(i = 7; i < args.nargs; i++) {
 				if(strlen(site.st_desc) + strlen(args.arg[i])
 				    + 2 > sizeof(site.st_desc))
 					break;
@@ -325,6 +469,76 @@ setsiteent(void)
 }
 
 
+int
+setsitegrp(char *grp)
+{
+	int found;
+
+	endsitegrp();
+	setsiteent();
+
+	group = list_init(grp, grp_comp, gfree, 0);
+	if(group == 0) {
+		cddbd_log(LOG_ERR, "Failed to allocate list head.");
+		quit(1);
+	}
+
+	found = 0;
+
+	while(getsiteent(SITE_GROUP) != NULL)
+		found++;
+
+	if(!found)
+		endsitegrp();
+
+	setsiteent();
+
+	return found;
+}
+
+
+void
+setsitemark(void)
+{
+	if((site_fp != NULL) && (sitemark != -1))
+		(void)fseek(site_fp, sitemark, SEEK_SET);
+}
+
+
+void
+getsitemark(void)
+{
+	if(site_fp == NULL)
+		return;
+
+	sitemark = ftell(site_fp);
+}
+
+
+void
+endsitegrp(void)
+{
+	if(group != 0) {
+		list_free(group);
+		group = 0;
+	}
+}
+
+
+void
+gfree(void *p)
+{
+	free(p);
+}
+
+
+int
+grp_comp(void *g1, void *g2)
+{
+	return(strcmp((char *)g1, (char *)g2));
+}
+
+
 void
 tfree(void *p)
 {
@@ -353,35 +567,53 @@ tfile_comp(void *t1, void *t2)
 
 
 void
-cddbd_rmt_op(int op, char *site, char *when)
+cddbd_rmt_op(int op, char *site, char *arg, char *arg2)
 {
 	int type;
 
-	if(op == RMT_OP_LOG)
+	/* T2D: check arguments */
+	switch(op) {
+	case RMT_OP_LOG:
 		type = SITE_LOG;
-	else
-		type = SITE_XMIT;
+		break;
 
-	if(strcmp(site, "all") && (getsitenam(site, type) == NULL)) {
+	default:
+		type = SITE_XMIT;
+		break;
+	}
+
+	if(strcmp(site, "all") && !setsitegrp(site) &&
+	    (getsitenam(site, type) == NULL)) {
 		cddbd_log(LOG_ERR,
 		    "Site \"%s\" is not defined in %s.", site, sitefile);
 		return;
 	}
 
-	(void)cddbd_clean_history();
-	endsiteent();
-
 	switch(op) {
-	case RMT_OP_TRANSMIT:
-		cddbd_transmit(site);
-		break;
-
 	case RMT_OP_LOG:
 		cddbd_remote_log(site);
 		break;
 
+	case RMT_OP_CMD:
+		cddbd_remote_cmd(site, arg, arg2);
+		break;
+
 	case RMT_OP_CATCHUP:
-		cddbd_catchup(site, when);
+		(void)cddbd_clean_history();
+		endsiteent();
+		cddbd_catchup(site, arg);
+
+		break;
+
+	case RMT_OP_PUT:
+		cddbd_put(site, arg);
+		break;
+
+	case RMT_OP_TRANSMIT:
+		(void)cddbd_clean_history();
+		endsiteent();
+		cddbd_transmit(site);
+
 		break;
 
 	default:
@@ -395,11 +627,11 @@ cddbd_rmt_op(int op, char *site, char *when)
 void
 cddbd_transmit(char *site)
 {
-	int f;
+	pid_t f;
 	int xcnt;
-	site_t *sp;
+	csite_t *sp;
 
-	if(strcmp(site, "all")) {
+	if(strcmp(site, "all") && !setsitegrp(site)) {
 		if((sp = getsitenam(site, SITE_XMIT)) != NULL)
 			cddbd_do_transmit(sp);
 	}
@@ -439,6 +671,7 @@ cddbd_transmit(char *site)
 				cddbd_do_transmit(sp);
 		}
 
+		endsitegrp();
 		endsiteent();
 
 		while(xcnt > 0) {
@@ -451,18 +684,19 @@ cddbd_transmit(char *site)
 
 
 void
-cddbd_do_transmit(site_t *sp)
+cddbd_do_transmit(csite_t *sp)
 {
 	int bps;
 	int xmit;
 	int fail;
 	int tcnt;
 	int files;
-	int encoding;
+	int flags= MF_NONE;
 	db_t *db;
 	time_t xtime;
 	FILE *fp;
 	lhead_t *li;
+	ct_key_t *key;
 	struct stat sbuf;
 	unsigned int discid;
 	char dir[CDDBBUFSIZ];
@@ -471,9 +705,16 @@ cddbd_do_transmit(site_t *sp)
 	char errstr[CDDBBUFSIZ];
 	char xmit_email[CDDBBUFSIZ];
 
+	if(sp->st_flags & ST_FLAG_NOXMIT) {
+		cddbd_log(LOG_ERR | LOG_XMIT,
+		    "Remote host %s is not configured for transmits.",
+		    sp->st_name);
+		return;
+	}
+
 	if(!cddbd_open_history(sp->st_name)) {
 		cddbd_log(LOG_ERR | LOG_XMIT,
-		    "Failed to open history file %s.", histfile);
+		    "Failed to open history file %s.", servfile[SF_HIST]);
 		return;
 	}
 
@@ -485,14 +726,16 @@ cddbd_do_transmit(site_t *sp)
 	}
 	else {
 		if(verbose)
-			cddbd_log(LOG_INFO, "Remote host %s up to date.",
+			cddbd_log(LOG_INFO, "Remote host %s is up to date.",
 			    sp->st_name);
 		return;
 	}
 
+	key = 0;
+
 	switch(sp->st_proto) {
 	case ST_PROTO_CDDBP:
-		if(!cddbp_open(sp->st_name, sp->st_port)) {
+		if(!cddbp_open(sp)) {
 			cddbd_log(LOG_ERR | LOG_XMIT,
 			    "Failed to open %s for file transmission.",
 			    sp->st_name);
@@ -509,6 +752,25 @@ cddbd_do_transmit(site_t *sp)
 		break;
 
 	case ST_PROTO_SMTP:
+		if(sp->st_flags & ST_FLAG_PWDLBL) {
+			if((key = getpasswd(sp->st_pwdlbl)) == NULL) {
+				cddbd_log(LOG_ERR | LOG_XMIT, "No password "
+				    "label \"%s\" found for site: %s",
+				    sp->st_pwdlbl, sp->st_name);
+
+				return;
+			}
+
+			flags = MF_MULTI;
+		}
+		else {
+			/* This is for supporting older sites. */
+			if(sp->st_flags & ST_FLAG_OMIME)
+				flags = MF_ENC;
+			else
+				flags = MF_MULTI;
+		}
+
 		if(!smtp_open()) {
 			cddbd_log(LOG_ERR | LOG_XMIT,
 			    "Failed to open SMTP for file transmission to %s.",
@@ -524,7 +786,7 @@ cddbd_do_transmit(site_t *sp)
 		}
 
 		cddbd_snprintf(xmit_email, sizeof(xmit_email), "%s@%s",
-		    sp->st_user, sp->st_name);
+		    sp->st_addr, sp->st_name);
 
 		break;
 
@@ -571,12 +833,12 @@ cddbd_do_transmit(site_t *sp)
 		}
 
 		/* Don't transmit the same file twice. */
-		if(list_find(li, (void *)(int)sbuf.st_ino) != 0) {
+		if(list_find(li, ino_t_as_ptr (sbuf.st_ino)) != 0) {
 			files--;
 			continue;
 		}
 
-		if(list_add_cur(li, (void *)(int)sbuf.st_ino) == 0) {
+		if(list_add_cur(li, ino_t_as_ptr (sbuf.st_ino)) == 0) {
 			cddbd_log(LOG_ERR, "Can't malloc linked list entry.");
 
 			if(!cddbd_close_history()) {
@@ -645,17 +907,16 @@ cddbd_do_transmit(site_t *sp)
 			cddbd_snprintf(subj, sizeof(subj), "cddb %s %08x",
 			    dir, discid);
 
-			if(db->db_flags & DB_REMAPPABLE)
-				encoding = CE_QUOTED_PRINT;
-			else
-				encoding = CE_7BIT;
-
 			if(!smtp_transmit(fp, subj, xmit_email, xmit_email,
-			    encoding, 0))
+			    admin_email, 0, flags, (int) sbuf.st_size, key))
 				fail++;
 			else {
 				xmit++;
 				tcnt += sbuf.st_size;
+
+				/* Be nice, wait an arbitrary amount of time. */
+				if(email_time > 0)
+					cddbd_delay(email_time);
 			}
 
 			break;
@@ -711,8 +972,6 @@ cddbd_do_transmit(site_t *sp)
 
 	switch(sp->st_proto) {
 	case ST_PROTO_CDDBP:
-		if(!fail)
-			(void)cddbp_update();
 		cddbp_close();
 		break;
 
@@ -745,12 +1004,12 @@ void
 cddbd_remote_log(char *site)
 {
 	int first;
-	site_t *sp;
+	csite_t *sp;
 
-	if(strcmp(site, "all")) {
+	if(strcmp(site, "all") && !setsitegrp(site)) {
 		if((sp = getsitenam(site, SITE_LOG)) != NULL &&
 		    sp->st_proto == ST_PROTO_CDDBP)
-			cddbd_do_remote_log(site, sp->st_port);
+			cddbd_do_remote_log(sp);
 
 		return;
 	}
@@ -765,22 +1024,23 @@ cddbd_remote_log(char *site)
 		if(!first)
 			printf("\n\n");
 
-		cddbd_do_remote_log(sp->st_name, sp->st_port);
+		cddbd_do_remote_log(sp);
 	}
 
+	endsitegrp();
 	endsiteent();
 }
 
 
 void
-cddbd_do_remote_log(char *site, int port)
+cddbd_do_remote_log(csite_t *sp)
 {
-	printf("Log statistics for %s:\n\n", site);
+	printf("Log statistics for %s:\n\n", sp->st_name);
 
-	if(!cddbp_open(site, port)) {
-		printf("Unknown host %s or host unavailable.\n", site);
+	if(!cddbp_open(sp)) {
+		printf("Unknown host %s or host unavailable.\n", sp->st_name);
 		cddbd_log(LOG_ERR | LOG_XMIT,
-		    "Failed to open %s for log acquisition.", site);
+		    "Failed to open %s for log acquisition.", sp->st_name);
 
 		return;
 	}
@@ -792,13 +1052,210 @@ cddbd_do_remote_log(char *site, int port)
 }
 
 
+
+
+void
+cddbd_remote_cmd(char *site, char *rcmd, char *plev)
+{
+	int first;
+	csite_t *sp;
+
+	if(strcmp(site, "all") && !setsitegrp(site)) {
+		if((sp = getsitenam(site, SITE_XMIT)) != NULL)
+			cddbd_do_remote_cmd(sp, rcmd, plev);
+
+		return;
+	}
+
+	for(first = 1;; first = 0) {
+		if((sp = getsiteent(SITE_XMIT)) == NULL)
+			break;
+
+		if(!first)
+			printf("\n\n");
+
+		cddbd_do_remote_cmd(sp, rcmd, plev);
+	}
+
+	endsitegrp();
+	endsiteent();
+}
+
+
+void
+cddbd_do_remote_cmd(csite_t *sp, char *rarg, char *plev)
+{
+	int lev;
+	char xmit_email[CDDBBUFSIZ];
+
+	if(verbose)
+		printf("Issuing command to %s:\n\n", sp->st_name);
+
+	if(strcmp(plev, "default"))
+		lev = atoi(plev);
+	else
+		lev = DEF_PROTO_LEVEL;
+
+	switch(sp->st_proto) {
+	case ST_PROTO_CDDBP:
+		if(!cddbp_open(sp)) {
+			printf("Unknown host %s or host unavailable.\n",
+			    sp->st_name);
+			return;
+		}
+
+		if(!cddbp_command(rarg, lev))
+			printf("Unable to issue command.\n");
+
+		cddbp_close();
+
+		break;
+
+	case ST_PROTO_SMTP:
+		if(!smtp_open()) {
+			printf("Unknown host %s or host unavailable.\n",
+			    sp->st_name);
+			return;
+		}
+
+		cddbd_snprintf(xmit_email, sizeof(xmit_email), "%s@%s",
+		    sp->st_addr, sp->st_name);
+
+		if(!smtp_command(sp, rarg, lev, xmit_email, admin_email))
+			printf("Unable to issue command.\n");
+
+		smtp_close();
+
+		printf("Email command sent to %s, return addr %s.\n",
+		    xmit_email, admin_email);
+
+		break;
+
+	default:
+		printf("Unsupported command protocol.\n");
+		break;
+	}
+}
+
+
+void
+cddbd_put(char *site, char *file)
+{
+	csite_t *sp;
+
+	if(strcmp(site, "all") && !setsitegrp(site)) {
+		if((sp = getsitenam(site, SITE_XMIT)) != NULL &&
+		    sp->st_proto == ST_PROTO_CDDBP)
+			cddbd_do_put(sp, file);
+
+		return;
+	}
+
+	for(;;) {
+		if((sp = getsiteent(SITE_XMIT)) == NULL)
+			break;
+
+		if(sp->st_proto != ST_PROTO_CDDBP)
+			continue;
+
+		cddbd_do_put(sp, file);
+	}
+
+	endsitegrp();
+	endsiteent();
+}
+
+
+void
+cddbd_do_put(csite_t *sp, char *file)
+{
+	if(!cddbp_open(sp)) {
+		cddbd_log(LOG_ERR | LOG_XMIT, "Failed to open %s for putting.",
+		    sp->st_name);
+
+		return;
+	}
+
+	(void)cddbd_put_file(sp->st_name, workdir, file);
+	cddbp_close();
+}
+
+
+int
+cddbd_put_file(char *site, char *dir, char *file)
+{
+	int ret;
+	FILE *fp;
+	DIR *dirp;
+	struct stat sbuf;
+	struct dirent *dp;
+	char pfile[CDDBBUFSIZ];
+	char *bfile;
+
+	cddbd_snprintf(pfile, sizeof(pfile), "%s/%s", dir, file);
+
+	if(stat(pfile, &sbuf)) {
+		cddbd_log(LOG_ERR | LOG_XMIT, "Can't stat file: %s (%d).",
+		    pfile, errno);
+
+		return 0;
+	}
+
+	if(S_ISDIR(sbuf.st_mode)) {
+		if((dirp = opendir(pfile)) == NULL) {
+			cddbd_log(LOG_ERR | LOG_XMIT,
+			    "Can't open directory: %s (%d).", pfile, errno);
+		}
+
+		while((dp = readdir(dirp)) != NULL) {
+			if(is_parent_dir(dp->d_name))
+				continue;
+
+			/* Transmit, and stop if there was a server error. */
+			if(cddbd_put_file(site, pfile, dp->d_name) == 0)
+				break;
+		}
+
+		closedir(dirp);
+
+		ret = 1;
+	}
+	else if(S_ISREG(sbuf.st_mode)) {
+		cddbd_log(LOG_INFO, "Putting %s on %s.", pfile, site);
+
+		if((fp = fopen(pfile, "r")) == NULL) {
+			cddbd_log(LOG_ERR | LOG_XMIT,
+			    "Can't open file for reading: %s (%d)",
+			    pfile, errno);
+
+			ret = -1;
+		}
+		else {
+			if((bfile = strrchr(file, '/')) == NULL)
+				bfile = file;
+
+			ret = cddbp_put(fp, bfile);
+			fclose(fp);
+		}
+	}
+	else {
+		cddbd_log(LOG_ERR | LOG_XMIT,
+		    "Can't put %s on %s: not a regular file", pfile, site);
+
+		ret = -1;
+	}
+
+	return ret;
+}
+
+
 void
 cddbd_catchup(char *site, char *when)
 {
-	site_t *sp;
-	site_t sitebuf;
+	csite_t *sp;
+	csite_t sitebuf;
 
-	if(strcmp(site, "all")) {
+	if(strcmp(site, "all") && !setsitegrp(site)) {
 		cddbd_do_catchup(site, when);
 		return;
 	}
@@ -806,21 +1263,18 @@ cddbd_catchup(char *site, char *when)
 	/* Catching up uses getsiteent, so be careful here. */
 	setsiteent();
 
-	if((sp = getsiteent(SITE_XMIT)) == NULL)
-		return;
-
-	sitebuf = *sp;
-
 	for(;;) {
+		if((sp = getsiteent(SITE_XMIT)) == NULL)
+			break;
+
 		sitebuf = *sp;
 		cddbd_do_catchup(sitebuf.st_name, when);
 
 		if(getsitenam(sitebuf.st_name, SITE_XMIT) == NULL)
 			break;
-		if((sp = getsiteent(SITE_XMIT)) == NULL)
-			break;
 	}
 
+	endsitegrp();
 	endsiteent();
 }
 
@@ -831,12 +1285,21 @@ cddbd_do_catchup(char *site, char *when)
 	int i;
 	time_t t;
 	DIR *dirp;
+	csite_t *sp;
 	struct stat sbuf;
 	struct dirent *dp;
 	char buf[CDDBBUFSIZ];
 	char cdir[CDDBBUFSIZ];
 	char file[CDDBBUFSIZ];
 
+	if((sp = getsitenam(site, SITE_XMIT)) != NULL &&
+	    (sp->st_flags & ST_FLAG_NOXMIT)) {
+		cddbd_log(LOG_ERR,
+		    "Remote host %s is not configured for transmits.",
+		    sp->st_name);
+		return;
+	}
+
 	/* Don't scan the database if they want to catch everything up. */
 	if(!strcmp("now", when)) {
 		t = time(0);
@@ -924,11 +1387,7 @@ cddbd_oc_history(char *name, int oflag)
 	unsigned int discid;
 	char buf[CDDBBUFSIZ];
 	char buf2[CDDBBUFSIZ];
-
-	if(histfile[0] == '\0') {
-		cddbd_log(LOG_ERR, "No history file defined.");
-		return 0;
-	}
+	char *thist;
 
 	/* Don't open if we already have, or close if we're not open. */
 	if(oflag && hist_opened)
@@ -937,32 +1396,31 @@ cddbd_oc_history(char *name, int oflag)
 	if(!oflag && !hist_opened)
 		return 0;
 
-	(void)cddbd_lock(lock_hist, 1);
+	(void)cddbd_lock(locks[LK_HIST], 1);
 
-	close(open(histfile, O_WRONLY | O_CREAT, file_mode));
+	close(open(servfile[SF_HIST], O_WRONLY | O_CREAT, file_mode));
 
-	(void)cddbd_fix_file(histfile, file_mode, uid, gid);
+	(void)cddbd_fix_file(servfile[SF_HIST], file_mode, uid, gid);
 
-	if((fp = fopen(histfile, "r")) == NULL) {
-		cddbd_log(LOG_ERR, "Can't open history file: %s.", histfile);
-		cddbd_unlock(lock_hist);
+	if((fp = fopen(servfile[SF_HIST], "r")) == NULL) {
+		cddbd_log(LOG_ERR, "Can't open history file: %s.",
+		    servfile[SF_HIST]);
+		cddbd_unlock(locks[LK_HIST]);
 		return 0;
 	}
 
-	cddbd_snprintf(thistfile, CDDBBUFSIZ, "%s_tmp", histfile);
+	thist = cddbd_mktemp();
 
-	if((tfp = fopen(thistfile, "w")) == NULL) {
+	if((tfp = fopen(thist, "w")) == NULL) {
 		cddbd_log(LOG_ERR, "Can't open tmp history file: %s.",
-		    thistfile);
+		    thist);
 
 		fclose(fp);
-		cddbd_unlock(lock_hist);
+		cddbd_unlock(locks[LK_HIST]);
 
 		return 0;
 	}
 
-	(void)cddbd_fix_file(thistfile, file_mode, uid, gid);
-
 	if(oflag) {
 		p = strdup(name);
 		if(p == NULL) {
@@ -991,7 +1449,7 @@ cddbd_oc_history(char *name, int oflag)
 				if(fputs(buf, tfp) == EOF) {
 					cddbd_log(LOG_ERR | LOG_XMIT,
 					    "Can't write tmp history file: %s.",
-					    thistfile);
+					    thist);
 					quit(QUIT_ERR);
 				}
 
@@ -1017,7 +1475,7 @@ cddbd_oc_history(char *name, int oflag)
 			if(fprintf(tfp, "%s/%08x\n", buf2, discid) == EOF) {
 				cddbd_log(LOG_ERR | LOG_XMIT,
 				    "Can't write tmp history file: %s.",
-				    thistfile);
+				    thist);
 				quit(QUIT_ERR);
 			}
 
@@ -1033,7 +1491,7 @@ cddbd_oc_history(char *name, int oflag)
 	if(!oflag && list_count(hist) > 0) {
 		if(fprintf(tfp, histstr, (char *)hist->lh_data) == EOF) {
 			cddbd_log(LOG_ERR | LOG_XMIT,
-			    "Can't write tmp history file: %s.", thistfile);
+			    "Can't write tmp history file: %s.", thist);
 			quit(QUIT_ERR);
 		}
 
@@ -1045,8 +1503,7 @@ cddbd_oc_history(char *name, int oflag)
 			if(fprintf(tfp, "%s/%08x\n", categlist[tf->tf_dir],
 			    tf->tf_discid) == EOF) {
 				cddbd_log(LOG_ERR | LOG_XMIT,
-				    "Can't write tmp history file: %s.",
-				    thistfile);
+				    "Can't write tmp history file: %s.", thist);
 				quit(QUIT_ERR);
 			}
 		}
@@ -1055,15 +1512,16 @@ cddbd_oc_history(char *name, int oflag)
 	fclose(fp);
 	fclose(tfp);
 
-	if(unlink(histfile) != 0 && errno != ENOENT) {
+	if(unlink(servfile[SF_HIST]) != 0 && errno != ENOENT) {
 		cddbd_log(LOG_ERR | LOG_XMIT,
-		    "Can't unlink %s (%d).", histfile, errno);
+		    "Can't unlink %s (%d).", servfile[SF_HIST], errno);
 		quit(QUIT_ERR);
 	}
 
-	if(cddbd_link(thistfile, histfile) != 0) {
+	if(cddbd_link(thist, servfile[SF_HIST]) != 0) {
 		cddbd_log(LOG_ERR | LOG_XMIT,
-		    "Can't link %s to %s (%d).", thistfile, histfile, errno);
+		    "Can't link %s to %s (%d).", thist, servfile[SF_HIST],
+		    errno);
 		quit(QUIT_ERR);
 	}
 
@@ -1076,12 +1534,8 @@ cddbd_oc_history(char *name, int oflag)
 		list_free(hist);
 	}
 
-	if(unlink(thistfile) != 0) {
-		cddbd_log(LOG_ERR | LOG_XMIT,
-		    "Warning: can't unlink %s (%d).", thistfile, errno);
-	}
-
-	cddbd_unlock(lock_hist);
+	cddbd_freetemp(thist);
+	cddbd_unlock(locks[LK_HIST]);
 
 	return 1;
 }
@@ -1158,16 +1612,11 @@ cddbd_count_history(char *site)
 	char buf[CDDBBUFSIZ];
 	char buf2[CDDBBUFSIZ];
 
-	if(histfile[0] == '\0') {
-		cddbd_log(LOG_ERR, "No history file defined.");
-		return -1;
-	}
+	(void)cddbd_lock(locks[LK_HIST], 1);
 
-	(void)cddbd_lock(lock_hist, 1);
-
-	if((fp = fopen(histfile, "r")) == NULL) {
+	if((fp = fopen(servfile[SF_HIST], "r")) == NULL) {
 		cddbd_log(LOG_ERR,
-		    "Can't open history file: %s.", histfile);
+		    "Can't open history file: %s.", servfile[SF_HIST]);
 		return -1;
 	}
 
@@ -1190,7 +1639,7 @@ cddbd_count_history(char *site)
 	}
 
 	fclose(fp);
-	cddbd_unlock(lock_hist);
+	cddbd_unlock(locks[LK_HIST]);
 
 	return cnt;
 }
@@ -1203,20 +1652,20 @@ cddbd_clean_history(void)
 	FILE *fp;
 	char buf[CDDBBUFSIZ];
 	char buf2[CDDBBUFSIZ];
+	csite_t *sp;
 
-	if(histfile[0] == '\0') {
-		cddbd_log(LOG_ERR, "No history file defined.");
-		return 0;
-	}
+	/* No history file yet. */
+	if(access(servfile[SF_HIST], F_OK))
+		return 1;
 
-	(void)cddbd_lock(lock_hist, 1);
+	(void)cddbd_lock(locks[LK_HIST], 1);
 
 	ret = 1;
 
 	for(fp = NULL; fp == NULL;) {
-		if((fp = fopen(histfile, "r")) == NULL) {
+		if((fp = fopen(servfile[SF_HIST], "r")) == NULL) {
 			cddbd_log(LOG_ERR,
-			    "Can't open history file: %s.", histfile);
+			    "Can't open history file: %s.", servfile[SF_HIST]);
 
 			ret = 0;
 			break;
@@ -1226,7 +1675,8 @@ cddbd_clean_history(void)
 		while(fgets(buf, sizeof(buf), fp) != NULL) {
 			/* If we found a nonexistent site, clear it. */
 			if(sscanf(buf, histstr, buf2) == 1 &&
-			    getsitenam(buf2, SITE_XMIT) == NULL) {
+			    ((sp = getsitenam(buf2, SITE_XMIT)) == NULL ||
+			    (sp->st_flags & ST_FLAG_NOXMIT))) {
 				fclose(fp);
 				fp = NULL;
 
@@ -1242,24 +1692,33 @@ cddbd_clean_history(void)
 	}
 
 	endsiteent();
-	cddbd_unlock(lock_hist);
+	cddbd_unlock(locks[LK_HIST]);
 
 	return ret;
 }
 
 
+/* return values: 0..error; 1..ok */
 int
 cddbd_merge_ulist(char *site, int merge)
 {
 	int ret;
-	site_t *sp;
+	csite_t *sp;
 
 	if(ulist_fp == NULL)
 		return 1;
 
+	/* Make sure this is not a group. */
+	if(setsitegrp(site)) {
+		endsitegrp();
+		return 0;
+	}
+
 	if(strcmp(site, "all")) {
 		if((sp = getsitenam(site, SITE_XMIT)) != NULL)
 			ret = cddbd_do_merge_ulist(sp->st_name, merge);
+		else ret= 0;
+		/* Note, BUG: ret would be undefined! */
 	}
 	else {
 		ret = 1;
@@ -1277,7 +1736,7 @@ cddbd_merge_ulist(char *site, int merge)
 
 
 	fclose(ulist_fp);
-	unlink(uhistfile);
+	cddbd_freetemp(uhist);
 	ulist_fp = NULL;
 	endsiteent();
 
@@ -1285,6 +1744,7 @@ cddbd_merge_ulist(char *site, int merge)
 }
 
 
+/* return values: 0..error; 1..ok */
 int
 cddbd_do_merge_ulist(char *site, int merge)
 {
@@ -1324,12 +1784,11 @@ int
 cddbd_open_ulist(void)
 {
 	if(ulist_fp == NULL) {
-		cddbd_snprintf(uhistfile, CDDBBUFSIZ, "%s.%05d",
-		    histfile, curpid);
+		uhist = cddbd_mktemp();
 
-		if((ulist_fp = fopen(uhistfile, "w+")) == NULL) {
+		if((ulist_fp = fopen(uhist, "w+")) == NULL) {
 			cddbd_log(LOG_ERR | LOG_UPDATE,
-			    "Can't open update-history file: %s.", uhistfile);
+			    "Can't open update-history file: %s.", uhist);
 			return 0;
 		}
 	}
@@ -1346,7 +1805,7 @@ cddbd_write_ulist(char *dir, char *disc)
 
 	if(fprintf(ulist_fp, "%s %s\n", dir, disc) == EOF) {
 		cddbd_log(LOG_ERR | LOG_UPDATE,
-		    "Can't write update-history file: %s.", uhistfile);
+		    "Can't write update-history file: %s.", uhist);
 		return 0;
 	}
 
@@ -1385,3 +1844,101 @@ copy_coord(char *buf, coord_t *cp)
 	cddbd_snprintf(buf, CDDBBUFSIZ, "%c%03d.%02d", cp->co_compass,
 	    cp->co_degrees, cp->co_minutes);
 }
+
+
+ct_key_t *
+getpasswd(char *lbl)
+{
+	int len;
+	int line;
+	FILE *fp;
+	ct_key_t *key;
+	struct stat sbuf;
+	char buf[CDDBBUFSIZ];
+	char buf2[CDDBBUFSIZ];
+
+	if(pwdfile[0] == '\0') {
+		cddbd_log(LOG_ERR, "No password file defined.");
+		return NULL;
+	}
+
+	if(stat(pwdfile, &sbuf) != 0) {
+		cddbd_log(LOG_ERR, "Can't stat password file: %s (%d)",
+		    pwdfile, errno);
+
+		return NULL;
+	}
+
+	if(sbuf.st_mode & (S_IRGRP | S_IWGRP | S_IROTH | S_IWOTH)) {
+		cddbd_log(LOG_ERR,
+		    "Insecure permissions (%04o) on password file: %s",
+		    (sbuf.st_mode & 0777), pwdfile);
+
+		return NULL;
+	}
+
+	if((fp = fopen(pwdfile, "r")) == NULL) {
+		cddbd_log(LOG_ERR, "Can't stat password file: %s (%d)",
+		    pwdfile, errno);
+
+		return NULL;
+	}
+
+	line = 0;
+	key = NULL;
+
+	while(fgets(buf, sizeof(buf), fp) != NULL) {
+		line++;
+
+		if(buf[0] == '#' || is_blank(buf, 0))
+			continue;
+
+		if(sscanf(buf, "%s", buf2) != 1) {
+			cddbd_log(LOG_ERR,
+			    "Illegal label %s in password file %s on line %d",
+			    lbl, pwdfile, line);
+
+			break;
+		}
+
+		len = strlen(buf2) - 1;
+
+		if(buf2[len] != ':') {
+			cddbd_log(LOG_ERR,
+			    "Syntax error in password file %s on line %d",
+			    pwdfile, line);
+		}
+
+		buf2[len] = '\0';
+
+		if(len > CDDBPLBLSIZ) {
+			cddbd_log(LOG_ERR,
+			    "Label %s too long in password file %s on line %d",
+			    lbl, pwdfile, line);
+
+			break;
+		}
+
+		if(strcmp(buf2, lbl))
+			continue;
+
+		if(sscanf(buf, "%*s%s", buf2) != 1) {
+			cddbd_log(LOG_ERR, "Missing password %s in password "
+			    "file %s on line %d", lbl, pwdfile, line);
+
+			break;
+		}
+
+		if((key = strtokey(buf2)) == 0) {
+			cddbd_log(LOG_ERR, "Illegal password for %s in password"
+			    " file %s on line %d", lbl, pwdfile, line);
+
+			key = NULL;
+		}
+
+		break;
+	}
+
+	fclose(fp);
+	return key;
+}
